from dls_utilpack.envvar import Envvar as DlsUtilpackEnvvar


class Envvar(DlsUtilpackEnvvar):
    """Class which covers environment variables, with default values."""

    XCHEMBKU_CONFIGFILE = "XCHEMBKU_CONFIGFILE"
