class Keywords:
    COMMAND = "xchembku_datafaces::keywords::command"
    PAYLOAD = "xchembku_datafaces::keywords::payload"


class Commands:
    EXECUTE = "xchembku_datafaces::commands::execute"


class Types:
    AIOHTTP = "xchembku_lib.xchembku_datafaces.aiohttp"
    DIRECT = "xchembku_lib.xchembku_datafaces.direct"
